from pathlib import Path

from pydantic import BaseModel, FilePath, validator

__all__ = ["SnapFile", "NewSnapFile"]


class SnapFile(BaseModel):
    path: FilePath

    @validator("path")
    def must_have_extension(cls, x: FilePath):
        if x.suffix != ".dcs":
            raise ValueError("must end in `.dcs`")
        return x


class NewSnapFile(BaseModel):
    path: Path

    @validator("path")
    def must_have_extension(cls, x: Path):
        if x.suffix != ".dcs":
            raise ValueError("must end in `.dcs`")
        return x

    @validator("path")
    def must_not_exist(cls, x: Path):
        if x.exists():
            raise ValueError("path already exists")
        return x

    @validator("path")
    def basedir_must_not_exist(cls, x: Path):
        if basedir(x).exists():
            raise ValueError(f"`{basedir(x)}` must not exist")
        return x

    @property
    def basedir(self):
        return basedir(self.path)


def basedir(x: Path):
    return x.parent / f"{x.stem}"



from __future__ import annotations

import shutil

from deciphon_core.cffi import ffi, lib
from deciphon_core.cseq import CSeqIter
from deciphon_core.error import DeciphonError
from deciphon_core.hmmfile import HMMFile
from deciphon_core.seq import SeqIter
from deciphon_core.snapfile import NewSnapFile

__all__ = ["Scan"]


class Scan:
    def __init__(self, hmm: HMMFile, seqit: SeqIter, snap: NewSnapFile):
        self._cscan = ffi.NULL
        self._hmm = hmm
        self._db = hmm.dbfile
        self._seqit = CSeqIter(seqit)
        self._snap = snap
        self._nthreads = 1
        self._port = 51371
        self._lrt_threshold = 10.0
        self._multi_hits = True
        self._hmmer3_compat = False

    @property
    def nthreads(self):
        return self._nthreads

    @nthreads.setter
    def nthreads(self, x: int):
        self._nthreads = x

    @property
    def port(self):
        return self._port

    @port.setter
    def port(self, x: int):
        self._port = x

    @property
    def lrt_threshold(self):
        return self._lrt_threshold

    @lrt_threshold.setter
    def lrt_threshold(self, x: float):
        self._lrt_threshold = x

    @property
    def multi_hits(self):
        return self._multi_hits

    @multi_hits.setter
    def multi_hits(self, x: bool):
        self._multi_hits = x

    @property
    def hmmer3_compat(self):
        return self._hmmer3_compat

    @hmmer3_compat.setter
    def hmmer3_compat(self, x: bool):
        self._hmmer3_compat = x

    def run(self):
        basedir = str(self._snap.basedir)
        if rc := lib.dcp_scan_run(self._cscan, basedir.encode()):
            raise DeciphonError(rc)

        archive = shutil.make_archive(basedir, "zip", base_dir=basedir)
        shutil.move(archive, self._snap.path)
        shutil.rmtree(self._snap.basedir)

    def open(self):
        self._cscan = lib.dcp_scan_new(self._port)
        if self._cscan == ffi.NULL:
            raise MemoryError()

        if rc := lib.dcp_scan_set_nthreads(self._cscan, self._nthreads):
            raise DeciphonError(rc)

        lib.dcp_scan_set_lrt_threshold(self._cscan, self._lrt_threshold)
        lib.dcp_scan_set_multi_hits(self._cscan, self._multi_hits)
        lib.dcp_scan_set_hmmer3_compat(self._cscan, self._hmmer3_compat)

        if rc := lib.dcp_scan_set_db_file(self._cscan, bytes(self._db.path)):
            raise DeciphonError(rc)

        it = self._seqit
        lib.dcp_scan_set_seq_iter(self._cscan, it.c_callback, it.c_self)

    def close(self):
        pass

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, *_):
        self.close()

    def __del__(self):
        if getattr(self, "_cscan", ffi.NULL) != ffi.NULL:
            lib.dcp_scan_del(self._cscan)



from __future__ import annotations

from typing import Optional

from deciphon_core.cffi import ffi, lib
from deciphon_core.error import DeciphonError
from deciphon_core.hmmfile import HMMFile

__all__ = ["Press"]


class HMMPress:
    def __init__(self, press: Press):
        self._press: Optional[Press] = press

    def press(self):
        assert self._press
        if rc := lib.dcp_press_next(self._press.cpress):
            raise DeciphonError(rc)
        self._press = None


class Press:
    def __init__(self, hmm: HMMFile):
        self._cpress = ffi.NULL
        self._hmm = hmm
        self._db = hmm.newdbfile
        self._nproteins: int = -1
        self._idx = -1

    @property
    def cpress(self):
        return self._cpress

    def open(self):
        self._cpress = lib.dcp_press_new()
        if self._cpress == ffi.NULL:
            raise MemoryError()

        hmmpath = bytes(self._hmm.path)
        dbpath = bytes(self._db.path)
        if rc := lib.dcp_press_open(self._cpress, hmmpath, dbpath):
            raise DeciphonError(rc)

    def close(self):
        if rc := lib.dcp_press_close(self._cpress):
            raise DeciphonError(rc)

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, *_):
        self.close()

    @property
    def nproteins(self) -> int:
        if self._nproteins == -1:
            self._nproteins: int = lib.dcp_press_nproteins(self._cpress)
        return self._nproteins

    def __len__(self):
        return self.nproteins

    def __iter__(self):
        return self

    def __next__(self):
        self._idx += 1
        if self._idx < self.nproteins:
            return HMMPress(self)
        raise StopIteration()

    def __del__(self):
        if getattr(self, "_cpress", ffi.NULL) != ffi.NULL:
            lib.dcp_press_del(self._cpress)



from pathlib import Path

from pydantic import BaseModel, FilePath, validator

from deciphon_core.dbfile import DBFile, NewDBFile

__all__ = ["HMMFile"]


class HMMFile(BaseModel):
    path: FilePath

    @validator("path")
    def must_have_extension(cls, x: FilePath):
        if x.suffix != ".hmm":
            raise ValueError("must end in `.hmm`")
        return x

    @property
    def _dbpath(self) -> Path:
        return self.path.parent / f"{self.path.stem}.dcp"

    @property
    def dbfile(self) -> DBFile:
        return DBFile(path=self._dbpath)

    @property
    def newdbfile(self) -> NewDBFile:
        return NewDBFile(path=self._dbpath)



from pathlib import Path

from pydantic import BaseModel, FilePath, validator

__all__ = ["DBFile", "NewDBFile"]


class DBFile(BaseModel):
    path: FilePath

    @validator("path")
    def must_have_extension(cls, x: FilePath):
        if x.suffix != ".dcp":
            raise ValueError("must end in `.dcp`")
        return x


class NewDBFile(BaseModel):
    path: Path

    @validator("path")
    def must_have_extension(cls, x: Path):
        if x.suffix != ".dcp":
            raise ValueError("must end in `.dcp`")
        return x

    @validator("path")
    def must_not_exist(cls, x: Path):
        if x.exists():
            raise ValueError("path already exists")
        return x



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from dataclasses import dataclass

__all__ = ["Seq", "SeqIter"]


@dataclass
class Seq:
    id: int
    name: str
    data: str


class SeqIter(metaclass=ABCMeta):
    @abstractmethod
    def __next__(self) -> Seq:
        ...
